import os
import pathlib
import subprocess
import json
import tempfile
import typing as T
from zenlog import log as logging
from accendino.localdeps import PackageManager
from accendino.utils import treatPackageDeps


class IToolChain:
    ''' Abstract toolchain '''

    def __init__(self, name, config):
        self.name = name
        self.description = name
        self.artifactRequires = {}
        self.config = config

    def packagesCheck(self, packageManager: PackageManager, artifacts: T.List[str], doInstall: bool) -> bool:
        '''
            checks that the toolchain have the necessary packages for the given toolchain artifacts
            @param packageManager: the manager for local package
            @param artifacts: a list of toolchain artifacts
            @param doInstall: tell if we should install missing packages
            @return if the operation was successful
        '''
        config = self.config
        shortName = f"{config.distribId}"
        longName = f"{config.distribId} {config.distribVersion}"
        if config.targetDistrib != config.distribId:
            # cross compiling adjusting the searched names
            shortName += f'->{config.targetDistrib}@{config.targetArch}'
            longName += f'->{config.targetDistrib}@{config.targetArch}'

        toCheck = []
        for artifact in artifacts:
            pkgRequires = self.artifactRequires.get(artifact)
            if pkgRequires:
                v = pkgRequires.get(longName, None)
                if v is None:
                    v = pkgRequires.get(shortName, [])
                if v:
                    toCheck += v

        logging.debug(f"packages to check from toolchain artifacts: {', '.join(toCheck)}")
        toInstall = packageManager.checkMissing(toCheck)
        if len(toInstall) and doInstall:
            return packageManager.installPackages(toInstall)

        return True

    def activate(self) -> bool:
        ''' Activates the toolchain, eventually doing some setup
            @return if the operation was successful
        '''
        return True

    def prepareItems(self) -> str:
        ''' returns some content to put in scripts to activate this toolchain
            @return the string content
        '''
        return ''

    def extraEnv(self, _artifacts) -> T.Dict[str, str]:
        ''' called to add some env variable for the toolchain for the given artifacts. Typically it
            sets variables like CC or CXX
            @param _artifacts: list of artifacts
            @returns some extra environment variable when building
        '''
        return {}

def computeEnvDiff(inputIter):
    '''
        computes the new env variables set after a call to VsDevCmd.bat, it parses content that
        looks like:
            var1=toto
            var2=tata
            ========
            **********************************************************************
            ** Visual Studio 2022 Developer Command Prompt v17.14.14
            ** Copyright (c) 2025 Microsoft Corporation
            **********************************************************************
            var1=toto
            var4=titi
            var3=tutu
            var2=tata
        and the output will be {'var3': 'tutu', 'var4': 'titi'}
    '''
    separator = '========'
    env1 = {}
    env2 = {}
    target = env1

    for l in inputIter:
        l = l.strip()
        if not l or l[0] in ('*',):
            continue

        if l == separator:
            target = env2
            continue

        var, value = l.split('=', 2)
        if var.lower() == 'path':
            var = 'PATH'
        target[var] = value

    diff = {}
    treatedKeys = []
    for k, v in env2.items():
        treatedKeys.append(k)
        if k not in env1 or env1[k] != v:
            diff[k] = v

    # sanity check should never print anything
    for k in env1:
        if k not in treatedKeys:
            logging.error(f"Warning: key {k} was removed by the script")

    return diff


KNOWN_VS_FLAVORS = ('msvc', 'clang')

class VsToolChain(IToolChain):
    ''' Toolchain using visual studio '''

    def __init__(self, config, flavor='msvc'):
        IToolChain.__init__(self, 'MsvcToolChain', config)
        self.artifactRequires = {
            'c': {
                'Windows': ['choco/vswhere|path/vswhere.exe'],
            },
            'c++': {
                'Windows': ['choco/vswhere|path/vswhere.exe'],
            },
        }
        self.installationPath = None
        self.installationName = None
        self.setvarsPath = None
        self.setvarsPs1Path = None
        self.flavor = flavor
        self.extraEnvMap = None



    def activate(self) -> bool:
        config = self.config
        if self.flavor == 'msvc':
            archsEquiv = {
                "x86_64": "x86.x64",
                "arm64": "ARM64"
            }

            component = f'Microsoft.VisualStudio.Component.VC.Tools.{archsEquiv.get(config.targetArch, "x86.x64")}'
        elif self.flavor == 'clang':
            component = 'Microsoft.VisualStudio.Component.VC.Llvm.Clang'
        else:
            logging.error(f'unknown VS flavor {self.flavor}')
            return False

        cmd = ['vswhere', '-latest', '-products', '*', '-requires', component, '-utf8', '-nocolor', '-format', 'json']
        logging.debug(f' * retrieving VS config by executing {" ".join(cmd)}')
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        content, _err = p.communicate()

        if p.returncode != 0:
            logging.error(f'error running vswhere, returnCode={p.returncode}')
            return False

        # returned output has an extra array struct it's like
        # [ {'instanceId': '8889553d', 'installDate': '2024-01-23T08:00:18Z', ... } ]
        #
        props = json.loads(content.decode('utf8'))[0]
        if not 'installationName' in props:
            logging.error('installationName not in vswhere output')
            return False

        self.installationName = props['installationName']

        if not 'installationPath' in props:
            logging.error('installationPath not in vswhere output')
            return False

        self.installationPath = pathlib.PurePath(props['installationPath'])
        self.setvarsPath = self.installationPath / "Common7" / "Tools" / "VsDevCmd.bat"
        self.setvarsPs1Path = self.installationPath / "Common7" / "Tools" / "Launch-VsDevShell.ps1"

        VSDEVCMD_ARCHS = {
            "x86_64": "amd64",
        }

        # run once the VsDevCmd.bat to catch the extra environment
        with tempfile.NamedTemporaryFile(suffix='.bat', mode='wt', encoding='utf-8', delete_on_close=False) as t:
            t.write(f'@set\n@echo ========\n@call "{str(self.setvarsPath)}" -arch={VSDEVCMD_ARCHS.get(config.targetArch, config.targetArch)}\n@set')
            t.flush()
            t.close()

            cmd = [ t.name ]
            logging.debug(f'running {" ".join(cmd)}')
            p = subprocess.run(cmd, stdout=subprocess.PIPE, encoding='utf-8')
            os.unlink(t.name)
            if p.returncode != 0:
                logging.error(f'error retrieving VS environment variables, errorCode={p.returncode}')
                return False

            self.extraEnvMap = computeEnvDiff(p.stdout.split('\n'))
            return True

    def extraEnv(self, _artifacts) -> T.Dict[str, str]:
        return self.extraEnvMap


class GccToolChain(IToolChain):
    ''' Toolchain with GCC '''

    def __init__(self, config):
        IToolChain.__init__(self, 'Gcc', config)
        self.artifactRequires = {
            'c': treatPackageDeps({
                'Debian|Ubuntu|Fedora|Redhat': ['gcc'],
                'Ubuntu->mingw@i686': ['gcc-mingw-w64-i686-posix'],
                'Ubuntu->mingw@x86_64': ['gcc-mingw-w64-x86-64-posix'],
                'Fedora->mingw@i686': ['mingw32-gcc', 'mingw32-crt'],
                'Fedora->mingw@x86_64': ['mingw64-gcc', 'mingw64-crt'],
            }),
            'c++': treatPackageDeps({
                'Debian|Ubuntu': ['g++'],
                'Fedora|Redhat': ['gcc-c++'],
            })
        }

    def extraEnv(self, artifacts) -> T.Dict[str, str]:
        ''' returns some extra environment variables
        '''
        ret = {}
        for artifact in artifacts:
            if artifact == 'c':
                ret['CC'] = 'gcc'
            elif artifact == 'c++':
                ret['CXX'] = 'g++'
        return ret


class ClangToolChain(IToolChain):
    ''' Toolchain with clang '''

    def __init__(self, config):
        IToolChain.__init__(self, 'Clang', config)
        self.artifactRequires = {
            'c': treatPackageDeps({
                'Debian|Ubuntu|Fedora|Redhat': ['clang']
            }),
            'c++': treatPackageDeps({
                'Debian|Ubuntu|Fedora|Redhat': ['clang']
            })
        }

    def extraEnv(self, artifacts) -> T.Dict[str, str]:
        ''' returns some extra environment variables
        '''
        ret = {}
        for artifact in artifacts:
            if artifact == 'c':
                ret['CC'] = 'clang'
            elif artifact == 'c++':
                ret['CXX'] = 'clang++'
        return ret

class MingwToolChain(IToolChain):
    ''' Toolchain with mingw '''

    def __init__(self, config):
        IToolChain.__init__(self, 'mingw', config)
        self.artifactRequires = {
            'c': treatPackageDeps({
                'Ubuntu->mingw@i686': ['gcc-mingw-w64-i686-posix'],
                'Debian->mingw@i686': ['gcc-mingw-w64-i686-posix'],
                'Ubuntu->mingw@x86_64': ['gcc-mingw-w64-x86-64-posix'],
                'Debian->mingw@x86_64': ['gcc-mingw-w64-x86-64-posix'],
                'Fedora->mingw@i686': ['mingw32-gcc', 'mingw32-crt'],
                'Fedora->mingw@x86_64': ['mingw64-gcc', 'mingw64-crt'],
            }),
            'c++': treatPackageDeps({
                'Ubuntu->mingw@i686': ['g++-mingw-w64-i686-win32'],
                'Debian->mingw@i686': ['g++-mingw-w64-i686-win32'],
                'Ubuntu->mingw@x86_64': ['g++-mingw-w64-x86-64-win32'],
                'Debian->mingw@x86_64': ['g++-mingw-w64-x86-64-win32'],
            })
        }
        self.config = config

    def extraEnv(self, artifacts) -> T.Dict[str, str]:
        VAR_ENV_PER_ARTIFACT = {
            'c': {
                'i686': {'CC': 'i686-w64-mingw32-gcc'},
                'x86_64': {'CC': 'x86_64-w64-mingw32-gcc'},
            },
            'c++': {
                'i686': {'CXX': 'i686-w64-mingw32-c++'},
                'x86_64': {'CXX': 'x86_64-w64-mingw32-c++'},
            }
        }

        ret = {}
        for artifact in artifacts:
            if artifact in VAR_ENV_PER_ARTIFACT:
                arch = self.config.targetArch
                if arch in VAR_ENV_PER_ARTIFACT[artifact]:
                    ret.update(VAR_ENV_PER_ARTIFACT[artifact][arch])

        return ret



class DefaultToolChain(IToolChain):
    ''' Toolchain that does some smart default depending '''

    def __init__(self, config):
        IToolChain.__init__(self, 'default', config)
        self.testObjs = []
        self.selectedObj = None

        if config.distribId == 'Windows':
            self.testObjs.append(VsToolChain(config, 'msvc'))
        else:
            self.testObjs.append(GccToolChain(config))
            self.testObjs.append(ClangToolChain(config))

    def packagesCheck(self, packageManager: PackageManager, artifacts: T.List[str], doInstall: bool) -> bool:
        logging.debug(' * autodetecting with default toolchain manager')
        for o in self.testObjs:
            logging.debug(f' * testing artifacts=[{", ".join(artifacts)}] against toolchain {o.name}')
            if o.packagesCheck(packageManager, artifacts, doInstall):
                logging.debug(f' * using {o.name} toolchain')
                self.selectedObj = o
                self.description = o.name
                return True
        return False

    def activate(self) -> bool:
        return self.selectedObj.activate()

    def prepareItems(self) -> str:
        return self.selectedObj.prepareItems()

    def extraEnv(self, artifacts) -> T.Dict[str, str]:
        return self.selectedObj.extraEnv(artifacts)



TOOLCHAINS = {
    'default': DefaultToolChain,
    'gcc': GccToolChain,
    'clang': ClangToolChain,
    'vs': VsToolChain,
    'mingw': MingwToolChain,
    'none': IToolChain,
}

def getToolchain(name: str, config):
    ret = None
    if name in TOOLCHAINS:
        ctor = TOOLCHAINS[name]
        ret = ctor(config)

    elif name.startswith('vs/'):
        flavor = name[3:]
        ret = VsToolChain(config, flavor)

    return ret




import unittest
from accendino.utils import NativePath

class Test(unittest.TestCase):


    def testNativePath(self):
        p = NativePath('tmp', 'ogon', prefix='--prefix=')
        self.assertEqual(p.prefix, '--prefix=')
        self.assertEqual(str(p), '--prefix=tmp/ogon')


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


import os
import shutil
import subprocess

from pathlib import Path
from zenlog import log as logging


class Source:
    ''' Generic code source '''

    def __init__(self, pkgs = {}):
        self.pkgDeps = pkgs


class LocalSource(Source):
    ''' Code source taken from a local directory that is either copied or symlinked '''

    def __init__(self, srcdir : str, do_symlink : bool = False) -> None:
        '''
            @param srcdir: source directory
            @param do_symlink: tells if the source tree must be symlinked or deep copied
        '''
        Source.__init__(self)
        self.srcdir = srcdir
        self.symlink = do_symlink

    def checkout(self, target_dir: str, _flog) -> bool:
        if os.path.exists(target_dir):
            if not self.symlink:
                if os.path.islink(target_dir):
                    return True
                logging.error(f"{target_dir} is a symlink, but we should do a hardcopy of the source tree" +
                              f" at {self.srcdir}. Please fix this by hand")
                return False

            logging.debug(f"==> refreshing dir {target_dir}")
            return True

        if self.symlink:
            logging.debug(f"==> linking {self.srcdir} to {target_dir}")
            os.symlink(os.path.abspath(self.srcdir), target_dir)
            return True

        try:
            logging.debug(f"==> copying {self.srcdir} to {target_dir}")
            shutil.copytree(src=self.srcdir, dst=target_dir)
            return True
        except Exception as e:
            logging.error(f"error copying tree: {e}")
            return False

class GitSource(Source):
    ''' Code source that is checked out from git '''

    def __init__(self, url: str, branch: str, depth: int = 1, shallow_submodules: bool = True,
                 recurse_submodules: bool = True) ->  None:
        '''
            @param url: URL of the git repo
            @param branch: branch to checkout
            @param depth: git --depth argument
            @param shallow_submodules: git --shallow-submodules argument
            @param recurse_submodules: git --recurse-submodules
        '''
        Source.__init__(self, {
            'Ubuntu|Debian': ['git'],
            'Fedora|Redhat': ['git'],
            'Windows': ['choco/git|path/git'],
        })

        self.url = url
        self.branch = branch
        self.depth = depth
        self.shallow_submodules = shallow_submodules
        self.recurse_submodules = recurse_submodules

    def checkout(self, target_dir: str, flog) -> bool:
        ''' '''
        if os.path.exists(target_dir):
            logging.debug(f"==> refreshing git dir {target_dir}")
            #cmd = ['git', 'pull']
            return True

        logging.debug(f"==> checking out repo in {target_dir}")
        cmd = ['git', 'clone', self.url, '-b', self.branch, target_dir]
        if self.depth:
            cmd += ['--depth', str(self.depth)]
        if self.shallow_submodules:
            cmd.append('--shallow-submodules')
        if self.recurse_submodules:
            cmd.append('--recurse-submodules')

        proc = subprocess.run(cmd, stdout=flog, stderr=flog)
        return proc.returncode == 0


class RemoteArchiveSource(Source):
    ''' Code source that is checked out from a remote location '''

    def __init__(self, url: str, saveAs: str = None, compression_method: str = 'guess') ->  None:
        '''
            @param url: URL of the archive repo
        '''
        Source.__init__(self, {
            'Ubuntu|Debian|Fedora|Redhat': ['curl'],
            'Windows': ['choco/curl|path/curl'],
        })

        if saveAs is None:
            pos = url.rfind("/")
            if pos == -1:
                logging.error(f'unable to extract filename from url {url}')
                raise NotImplementedError()
            self.saveAs = url[pos+1:]
        else:
            self.saveAs = saveAs

        knownCompressions = {
            'tar.gz': (['tar', 'xzf'],
                {
                    'Ubuntu|Debian|Fedora|Redhat': ['tar', 'gzip'],
                    'Windows': ['choco/7z|path/7z']
                },

            ),
            'tar': (['tar', 'xf'],
                {
                    'Ubuntu|Debian|Fedora|Redhat': ['tar'],
                    'Windows': ['choco/7zip|path/7z']
                },
            ),
            'zip': (['unzip', '-o'],
                {
                    'Ubuntu|Debian|Fedora|Redhat': ['unzip'],
                    'Windows': ['choco/7z|path/7z']
                },
            ),
            '7z': (['7z', 'x'],
                {
                    'Ubuntu|Debian|Fedora|Redhat': ['unzip'],
                    'Windows': ['choco/7z|path/7z']
                },
            ),
        }

        self.url = url
        self.compression = compression_method

        compProps = None
        if compression_method == 'guess':
            for ext, props in knownCompressions.items():
                if url.endswith('.' + ext):
                    compProps = props
                    break

            if compProps is None:
                logging.error(f'unable to guess compression method from url {url}')
                raise NotImplementedError()

        else:
            compProps = knownCompressions.get(compression_method, None)

        if compression_method and compProps is None:
            logging.error(f'unknown compression method {compression_method}')
            raise NotImplementedError()

        self.decompressCmd = None
        if compProps:
            # update native package deps
            for k, v in compProps[1].items():
                if k in self.pkgDeps:
                    self.pkgDeps[k] += v
                else:
                    self.pkgDeps[k] = [ v ]

            self.decompressCmd = compProps[0]
        else:
            if compression_method:
                logging.error(f'unknown compression method {compression_method}')
                raise NotImplementedError()

    def decompress(self, target_dir, sourcePath, flog):
        if self.decompressCmd is None:
            return True

        extractCmd = self.decompressCmd + [ str(sourcePath) ]
        logging.debug(f'running {" ".join(extractCmd)} in {target_dir}')
        proc = subprocess.run(extractCmd, cwd=target_dir, stdout=flog, stderr=flog)
        if proc.returncode != 0:
            logging.error(f'error extracting {sourcePath} in {target_dir}')
            return False
        return True


    def checkout(self, target_dir, flog) -> bool:
        archiveDir = target_dir / '..' / '..' / 'archives'
        if not os.path.exists(archiveDir):
            os.makedirs(archiveDir, exist_ok=True)

        saveAsPath = Path(archiveDir / self.saveAs).resolve()
        retrieveCmd = ['curl', '-s', '-L', self.url, '-o', str(saveAsPath)]

        if os.path.exists(saveAsPath):
            retrieveCmd += ['-z', str(saveAsPath)]

        logging.debug(f'running {" ".join(retrieveCmd)}')
        proc = subprocess.run(retrieveCmd, stdout=flog, stderr=flog)
        if proc.returncode != 0:
            logging.error(f'error retrieving {self.url}')
            return False

        return self.decompress(target_dir, saveAsPath, flog)



import sys
import os

def findFirstExistingPath(cands):
    for cand in cands:
        if os.path.exists(cand):
            return cand
    return None


class AccendinoPlatform:
    ''' '''
    def __init__(self):
        self.isWindows = sys.platform.startswith("win")
        if self.isWindows:
            self.msys2path = findFirstExistingPath(['c://tools/msys64/msys2_shell.cmd', 'c://msys64/msys2_shell.cmd'])
        else:
            self.msys2path = None

accendinoPlatform = AccendinoPlatform()



import json
from accendino.toolchain import computeEnvDiff

v = '''[
  {
    "instanceId": "8889553d",
    "installDate": "2024-01-23T08:00:18Z",
    "installationName": "VisualStudio/17.14.14+36429.23",
    "installationPath": "C:\\\\Program Files\\\\Microsoft Visual Studio\\\\2022\\\\Community",
    "installationVersion": "17.14.36429.23",
    "productId": "Microsoft.VisualStudio.Product.Community",
    "productPath": "C:\\\\Program Files\\\\Microsoft Visual Studio\\\\2022\\\\Community\\\\Common7\\\\IDE\\\\devenv.exe",
    "state": 4294967295,
    "isComplete": true,
    "isLaunchable": true,
    "isPrerelease": false,
    "isRebootRequired": false,
    "displayName": "Visual Studio Community 2022",
    "description": "IDE puissant, gratuit pour les étudiants, les contributeurs open source et les particuliers",
    "channelId": "VisualStudio.17.Release",
    "channelUri": "https://aka.ms/vs/17/release/channel",
    "enginePath": "C:\\\\Program Files (x86)\\\\Microsoft Visual Studio\\\\Installer\\\\resources\\\\app\\\\ServiceHub\\\\Services\\\\Microsoft.VisualStudio.Setup.Service",
    "installedChannelId": "VisualStudio.17.Release",
    "installedChannelUri": "https://aka.ms/vs/17/release/channel",
    "releaseNotes": "https://docs.microsoft.com/en-us/visualstudio/releases/2022/release-notes-v17.14#17.14.14",
    "resolvedInstallationPath": "C:\\\\Program Files\\\\Microsoft Visual Studio\\\\2022\\\\Community",
    "thirdPartyNotices": "https://go.microsoft.com/fwlink/?LinkId=661288",
    "updateDate": "2025-09-12T19:56:45.150353Z",
    "catalog": {
      "buildBranch": "d17.14",
      "buildVersion": "17.14.36429.23",
      "id": "VisualStudio/17.14.14+36429.23",
      "localBuild": "build-lab",
      "manifestName": "VisualStudio",
      "manifestType": "installer",
      "productDisplayVersion": "17.14.14",
      "productLine": "Dev17",
      "productLineVersion": "2022",
      "productMilestone": "RTW",
      "productMilestoneIsPreRelease": "False",
      "productName": "Visual Studio",
      "productPatchVersion": "14",
      "productPreReleaseMilestoneSuffix": "1.0",
      "productSemanticVersion": "17.14.14+36429.23",
      "requiredEngineVersion": "3.14.2086.54749"
    },
    "properties": {
      "campaignId": "2030:084e6fb6-0816-4291-b1cf-71d75f4176bc",
      "channelManifestId": "VisualStudio.17.Release/17.14.14+36429.23",
      "nickname": "",
      "setupEngineFilePath": "C:\\\\Program Files (x86)\\\\Microsoft Visual Studio\\\\Installer\\\\setup.exe"
    }
  }
]'''


#obj = json.loads(v)
#print(obj)

if __name__ == '__main__':
    v = open('cmdOutput.txt', 'r').readlines()

    diff = computeEnvDiff(v)



import os
import subprocess
import typing as T
import re

from zenlog import log as logging
from accendino.utils import findInPATH
from accendino.platform import accendinoPlatform


class PackageManagerBase:
    ''' base package manager '''

    def __init__(self, name: str=None) -> None:
        self.name = name
        self.allPackages = {}
        self.debug = True
        self.canInstall = True

    def checkMissing(self, packages: T.List[str]) -> T.List[str]:
        ret = []

        logging.debug(f" * {self.name}: checking required {len(packages)} package(s) on system")

        for p in packages:
            if p not in self.allPackages:
                logging.debug(f"   {p}: missing")
                ret.append(p)

        return ret

    def installPackages(self, _packages: T.List[str]) -> bool:
        logging.error(f"{self.name}, not implemented")
        return False




class DpkgManager(PackageManagerBase):
    ''' dpkg based package manager '''

    def __init__(self) -> None:
        PackageManagerBase.__init__(self, "dpkg")

        pack_re = re.compile(r'ii[^\w]+([^ ]+)[^\w]+([^ ]+)')
        for l in subprocess.Popen(['dpkg', '-l'], stdout=subprocess.PIPE, bufsize=1024).stdout.readlines():
            matches = pack_re.match(l.decode('utf8'))
            if not matches:
                continue

            pkgName = matches.group(1)
            pos = pkgName.find(':')
            if pos != -1:
                pkgName = pkgName[0:pos]

            version = matches.group(2)
            self.allPackages[pkgName] = version

        logging.debug(f" * {self.name} package manager: got {len(self.allPackages)} installed packages")

    def installPackages(self, packages: T.List[str]) -> bool:
        logging.debug(f" * {self.name}, installing missing packages: {' '.join(packages)}")

        cmd = f"apt-get install -y --no-install-recommends {' '.join(packages)}"
        if os.getuid() != 0:
            cmd = "sudo " + cmd

        return os.system(cmd) == 0


class RpmManager(PackageManagerBase):
    ''' rpm based package manager '''

    def __init__(self) -> None:
        PackageManagerBase.__init__(self, "rpm")

        for l in subprocess.Popen(['rpm', '-qa', '--qf', '%{NAME} %{VERSION}\\n'], stdout=subprocess.PIPE, bufsize=1024) \
                            .stdout.readlines():
            tokens = l.decode('utf-8').strip().split(' ', 2)
            self.allPackages[tokens[0]] = tokens[1]

        logging.debug(f" * {self.name} package manager: got {len(self.allPackages)} installed packages")

    def installPackages(self, packages: T.List[str]) -> bool:
        logging.debug(f" * {self.name}, installing missing packages: {' '.join(packages)}")

        cmd = f"dnf -y install {' '.join(packages)}"
        if os.getuid() != 0:
            cmd = "sudo " + cmd

        return os.system(cmd) == 0

class BrewManager(PackageManagerBase):
    ''' brew based package manager '''

    def __init__(self) -> None:
        PackageManagerBase.__init__(self, "brew")
        for l in subprocess.Popen(['brew', 'list', '--formulae', '--versions'], stdout=subprocess.PIPE, bufsize=1024) \
                            .stdout.readlines():
            tokens = l.decode('utf-8').strip().split(' ', 2)
            self.allPackages[tokens[0]] = tokens[1]

        logging.debug(f" * {self.name} package manager: got {len(self.allPackages)} installed packages")


    def installPackages(self, packages: T.List[str]) -> bool:
        logging.debug(f" * {self.name}, installing missing packages: {' '.join(packages)}")

        cmd = f"brew install {' '.join(packages)}"
        return os.system(cmd) == 0


class InPathSubManager(PackageManagerBase):
    ''' package manager that checks for programs in PATH '''

    def __init__(self) -> None:
        PackageManagerBase.__init__(self, "inPath")
        self.canInstall = False

    def checkMissing(self, packages: T.List[str]) -> T.List[str]:
        ret = []
        for p in packages:
            if findInPATH(p) is None:
                ret.append(p)

        return ret

    def installPackages(self, packages: T.List[str]) -> bool:
        logging.error(f'InPathSubManager can\'t install {" ".join(packages)}')
        return False

class ChocoManager(PackageManagerBase):
    ''' Chocolatey based package manager '''

    def __init__(self, chocoPath: str) -> None:
        PackageManagerBase.__init__(self, "chocolatey")
        self.chocoPath = chocoPath

        for l in subprocess.Popen([chocoPath, 'list', '--no-color', '-r'], stdout=subprocess.PIPE, bufsize=1024) \
                            .stdout.readlines():
            l = l.decode('utf-8').strip()

            # output looks like:
            #     chocolatey|2.5.1
            #     nasm|2.16.3
            if not l:
                continue

            tokens = l.split('|', 2)
            self.allPackages[tokens[0]] = tokens[1]

        logging.debug(f" * {self.name} package manager: got {len(self.allPackages)} installed packages")

    def installPackages(self, packages: T.List[str]) -> bool:
        logging.debug(f" * {self.name}, installing missing packages: {' '.join(packages)}")

        cmd = [self.chocoPath, "install", "-y"] + packages
        p = subprocess.run(cmd)
        return p.returncode == 0


class PacmanManager(PackageManagerBase):
    ''' Pacman based package manager for Arch or msys2 '''

    def __init__(self, name: str = 'pacman') -> None:
        PackageManagerBase.__init__(self, name)
        for l in self.executePipe(['pacman', '-Q']).stdout.readlines():
            tokens = l.decode('utf-8').strip().split(' ', 2)
            self.allPackages[tokens[0]] = tokens[1]

        logging.debug(f" * {self.name} package manager: got {len(self.allPackages)} installed packages")

    def executePipe(self, cmd: T.List[str]):
        logging.debug(f"executing {' '.join(cmd)}")
        return subprocess.Popen(cmd, stdout=subprocess.PIPE, bufsize=1024)

    def installPackages(self, packages: T.List[str]) -> bool:
        logging.debug(f" * {self.name}, installing missing packages: {' '.join(packages)}")

        cmd = []
        if not accendinoPlatform.isWindows:
            if os.getuid() != 0:
                cmd.append("sudo")

        cmd += ['pacman', '-S', '--noprogressbar', '--noconfirm', ' '.join(packages)]
        p = self.executePipe(cmd)
        out, _err = p.communicate()

        if p.returncode != 0:
            logging.error(f"error installing {' '.join(packages)}, errorCode={p.returncode}: {out}")
            return False

        return True


class Msys2Manager(PacmanManager):
    ''' '''
    def __init__(self, path) -> None:
        self.msysShellPath = path
        PacmanManager.__init__(self, "msys2")

    def executePipe(self, cmd: T.List[str]):
        newCmd = [self.msysShellPath, '-defterm', '-no-start', '-mingw64', '-here', '-c', " ".join(cmd)]
        return PacmanManager.executePipe(self, newCmd)

class PkgManager(PackageManagerBase):
    ''' pkg based package manager for FreeBSD '''

    def __init__(self) -> None:
        PackageManagerBase.__init__(self, "pkg")

        for l in subprocess.Popen(['pkg', 'info'], stdout=subprocess.PIPE, bufsize=1024) \
                    .stdout.readlines():
            v = l.decode('utf-8').strip().split(' ', 2)[0]
            pos = v.rfind('-')
            name = v[0:pos]
            version = v[pos+1:]
            self.allPackages[name] = version

        logging.debug(f" * ${self.name} package manager: got {len(self.allPackages)} installed packages")

    def installPackages(self, packages: T.List[str]) -> bool:
        logging.debug(f" * {self.name}, installing missing packages: {' '.join(packages)}")

        cmd = f"pkg install -y {' '.join(packages)}"
        if os.getuid() != 0:
            cmd = "sudo " + cmd

        return os.system(cmd) == 0

class PackageManager(PackageManagerBase):

    def __init__(self, name, managers):
        PackageManagerBase.__init__(self, name)
        self.managers = managers

    def checkMissing(self, packages: T.List[str]) -> T.List[str]:
        packagesPerManager = {}
        ret = []

        for p in packages:
            alternatives = p.split('|')

            if len(alternatives) > 1:
                #
                # handle syntax with "choco/nasm|nuget/nasm2|path/nasm.exe" that would check for
                #  nasm with choco, then nasm2 with nuget and finally nasm.exe in the path
                #
                found = False
                cand = None

                for alter in alternatives:
                    tokens = alter.strip().split('/', 2)
                    if len(tokens) == 1:
                        # no sub manager set, using default
                        manager = ''
                        package = tokens[0]
                    else:
                        (manager, package) = tokens

                    managerObj = self.managers.get(manager, None)
                    if managerObj:
                        if len(managerObj.checkMissing([package])) == 0:
                            found = True
                            break
                        if managerObj.canInstall and not cand:
                            cand = alter

                if not found:
                    if cand:
                        logging.debug(f'{p} not found, pushing {cand} for installation')
                        ret.append(cand)
                    else:
                        logging.error(f'can\'t find any alternative for {p}')
                        return None

            else:
                tokens = p.strip().split('/', 2)
                if len(tokens) == 1:
                    package = tokens[0]
                    manager = ''
                else:
                    (manager, package) = tokens

                if manager in packagesPerManager:
                    packagesPerManager[manager].append(package)
                else:
                    packagesPerManager[manager] = [package]

        helpMsg = {
            '': "no default manager set",
            'choco': "some choco packages where requested but Chocolatey is not installed",
            'msys2': "some msys2 packages where requested but msys2 is not installed, perhaps try 'choco install msys2'",
        }

        for managerName, pkgs in packagesPerManager.items():
            managerObj = self.managers.get(managerName, None)
            # no sub manager set, using default
            if managerObj is None:
                msg = helpMsg.get(managerName, None)
                if msg:
                    logging.error(msg)
                else:
                    logging.error(f"no manager {managerName} set to install {pkgs}")
                return None

            toInstall = managerObj.checkMissing(pkgs)

            if toInstall and not managerObj.canInstall:
                logging.error(f"missing items for manager {managerName}, but it can't install {pkgs}")
                return None

            for p in toInstall:
                if not managerName:
                    ret.append(p)
                else:
                    ret.append(f"{managerName}/{p}")

        return ret

    def installPackages(self, packages: T.List[str]) -> bool:
        logging.debug(f" * {self.name}, installing missing packages: {' '.join(packages)}")

        packagesPerManager = {}
        for p in packages:
            tokens = p.split('/', 2)
            if len(tokens) == 1:
                managerName = ''
                package = tokens[0]
            else:
                (managerName, package) = tokens

            if managerName not in self.managers:
                logging.error(f"{managerName} not registered")
                return False

            if managerName in packagesPerManager:
                packagesPerManager[managerName].append(p)
            else:
                packagesPerManager[managerName] = [ package ]


        for managerName, pkgs in packagesPerManager.items():
            visualName = managerName and managerName or 'default'
            logging.info(f'installing [{" ".join(pkgs)}] on {visualName} package manager')
            manager = self.managers[managerName]
            if not manager.installPackages(pkgs):
                return False

        return True



def getPkgManager(distribId, packagesToCheck):
    managers = {'path': InPathSubManager()}
    name = 'unknown'
    if distribId in ('Ubuntu', 'Debian', ):
        name = 'dpkg'
        managers[''] = DpkgManager()
    elif distribId in ('Fedora', 'RedHat', ):
        name = 'rpm'
        managers[''] = RpmManager()
    elif distribId in ('Windows',):
        name = 'windows'
        chocoPath = findInPATH("choco.exe")
        if chocoPath:
            managers['choco'] = ChocoManager(chocoPath)

        if accendinoPlatform.msys2path:
            managers['msys2'] = Msys2Manager(accendinoPlatform.msys2path)
        packagesToCheck.append('path/powershell.exe')

    elif distribId in ('Darwin',):
        name = 'brew'
        managers[''] = BrewManager()
    elif distribId in ('FreeBSD',):
        name = 'pkg'
        managers[''] = PkgManager()
    elif distribId in ('Arch',):
        name = 'pacman'
        managers[''] = PacmanManager()

    return PackageManager(name, managers)



import os
import subprocess
import pickle
import pathlib
import typing as T

from zenlog import log as logging
from accendino.sources import Source
from accendino.utils import mergePkgDeps, treatPackageDeps, doMingwCrossDeps, RunInShell, as_msys2_path, \
    getArchLibDir


class BuildStepDump:
    def __init__(self):
        self.gitCommit = None
        self.env = None
        self.args = None

    def __eq__(self, other) -> bool:
        if self.gitCommit != other.gitCommit:
            return False

        # check environment
        for k in self.env.keys():
            if not k in other.env:
                return False

            if self.env[k] != other.env.get(k, None):
                return False

        if len(self.args) != len(other.args):
            return False

        # check commands
        i = 0
        for (cmds1, dir1, comment1) in self.args:
            (cmds2, dir2, comment2) = other.args[i]

            if dir1 != dir2 or comment1 != comment2:
                return False

            if isinstance(cmds1, RunInShell):
                cmds1 = cmds1.expand()

            if isinstance(cmds2, RunInShell):
                cmds1 = cmds1.expand()

            if len(cmds1) != len(cmds2):
                return False

            j = 0
            for s1 in cmds1:
                # check commands by stringify them for a correct NativePath comparison
                if not isinstance(s1, str):
                    s1 = str(s1)

                s2 = cmds2[j]
                if not isinstance(s2, str):
                    s2 = str(s2)

                if s1 != s2:
                    return False
                j += 1

            i += 1

        return True

PREPARE_DUMP_FILE = 'accendino.prepared'
BUILT_FILE = 'accendino.built'


class DepsBuildArtifact:
    ''' basic build artifact only for dependencies and platform packages '''

    def __init__(self, name, deps=[], provides=[], pkgs={}, toolchainArtifacts='c') -> None:
        '''
            @param name: name of the build artifact
            @param deps: list of dependencies to other build artifacts
            @param provides: list of provided build artifacts
            @param pkgs: required platform packages
            @param toolchainArtifacts: artifacts that we need from the toolchain
        '''
        self.name = name
        self.deps = deps
        self.provides = provides
        self.pkgs = treatPackageDeps(pkgs)
        self.prepareStateFile = None
        self.builtFile = None
        if isinstance(toolchainArtifacts, str):
            self.toolchainArtifacts = toolchainArtifacts.split(',')
        else:
            self.toolchainArtifacts = toolchainArtifacts

    def init(self, _config) -> bool:
        return True

    def checkout(self, _config) -> bool:
        return True

    def prepare(self, _config) -> bool:
        return True

    def build(self, _config) -> bool:
        return True

    def __str__(self) -> str:
        return f"<{self.name}>"

WIN_PREPARE_SCRIPT = 'prepare.ps1'
WIN_BUILD_SCRIPT = 'build.ps1'

class BuildArtifact(DepsBuildArtifact):
    ''' general build artifact '''

    def __init__(self, name: str, deps, srcObj: Source, extraEnv={}, provides=[], pkgs={}, prepare_cmds = [], build_cmds=[],
                 toolchainArtifacts='c', skipToolchainEnv=False) -> None:
        '''
            'commands list' are list of tuple(command, running_directory, command_description).
                * command is the list passed to subprocess.execute().
                * running_directory where to run the command (you can use {srcdir}, {builddir} they will be
                  replaced by their values).
                * command_description is used for logging and error reporting

            @param name: name of the build artifact
            @param deps: list of dependencies to other build artifacts
            @param srcObj: Source object for checking out code
            @param extraEnv: extra environment variable to use when running commands
            @param provides: list of provided build artifacts
            @param pkgs: required platform packages
            @param prepare_cmds: a command list of commands to run to prepare the build tree.
            @param build_cmds: a command list of commands to run to do the build and install
            @param toolchainArtifacts: artifacts that we need from the toolchain
            @param skipToolchainEnv: do not grab env variable from the toolchain
        '''
        if srcObj:
            pkgs = mergePkgDeps(pkgs, srcObj.pkgDeps)

        DepsBuildArtifact.__init__(self, name, deps, provides, pkgs, toolchainArtifacts)
        self.srcObj = srcObj
        self.sourceDir = None
        self.buildDir = None
        self.extraEnv = extraEnv
        self.logFile = None
        self.prepare_cmds = prepare_cmds[:]
        self.build_cmds = build_cmds[:]
        self.parallelJobs = True
        self.needsMsys2 = False
        self.skipToolchainEnv = skipToolchainEnv

    def _updatePATHlike(self, config, env: T.Dict[str, str], key: str, preExtra: T.List[str] = [],
                        postExtra: T.List[str] = [], sep: str = ':') -> None:
        ''' updates an env variable that is like PATH or PKG_CONFIG_PATH '''
        value = env.get(key, None)
        if value:
            l = value.split(sep)
        else:
            l = []

        ret = []
        for item in preExtra + l + postExtra:
            item = self._expandConfigInString(item, config)
            if item not in ret:
                ret.append(item)

        if len(ret) > 0:
            env[key] = sep.join(ret)


    def _createEnvFileUnix(self, env: T.Dict[str, str], keys: T.List[str]) -> None:
        with open(self.buildDir / 'setEnv.sh', 'wt', encoding='utf8') as f:
            f.write(f'# environment variables for artifact {self.name}\n')
            for k in keys:
                f.write(f'export {k}="{env[k]}"\n')

    def _pushPowerShellEnv(self, f, env: T.Dict[str, str], keys: T.List[str]) -> None:
        f.write(f'# environment variables for artifact {self.name}\n')
        for k in keys:
            f.write(f"$env:{k} = '{env[k]}'\n")

        f.write('\n')

    def _createEnvFileWin32(self, env: T.Dict[str, str], keys: T.List[str]) -> None:
        with open(self.buildDir / 'setEnv.ps1', 'wt', encoding='utf8') as f:
            self._pushPowerShellEnv(f, env, keys)

    def _updateEnvMap(self, config, base: T.Dict[str, str], extra: T.Dict[str, str]) -> T.Dict[str, str]:
        for k, v in extra.items():
            if not isinstance(v, str):
                v = str(v)

            if k == 'PATH':
                self._updatePATHlike(config, base, 'PATH', v.split(os.pathsep), sep=os.pathsep)
            elif k == 'PKG_CONFIG_PATH':
                self._updatePATHlike(config, base, 'PKG_CONFIG_PATH', v.split(':'))
            else:
                base[k] = v

    def _computeEnv(self, config, extra: T.Dict[str, str], createEnvFile: bool=False) -> T.Tuple[T.Dict[str, str], T.List[str]]:
        r = os.environ.copy()
        xkeys = []

        if self.needsMsys2:
            r['MSYS2_PATH_TYPE'] = 'inherit'
            xkeys.append('MSYS2_PATH_TYPE')

        # start by updating the base environment with what's comes from the toolchain
        if not self.skipToolchainEnv:
            toolchainEnv = config.toolchainObj.extraEnv(self.toolchainArtifacts)
            if toolchainEnv:
                self._updateEnvMap(config, r, toolchainEnv)
                xkeys += toolchainEnv.keys()

        # add a PKG_CONFIG_PATH
        xkeys.append('PKG_CONFIG_PATH')
        pkg_config_path = '{prefix_posix}/{libdir}/pkgconfig'

        archLibDir = getArchLibDir(config.targetDistrib, config.targetArch)
        if archLibDir:
            pkg_config_path += ':{prefix_posix}/{libdir}/' + archLibDir + '/pkgconfig'

        xx = {
            'PKG_CONFIG_PATH': pkg_config_path,
        }

        # adds the target bin directory to the PATH
        if not config.crossCompilation:
            xx['PATH'] = pathlib.PurePath(config.prefix, 'bin')
            xkeys.append('PATH')

        self._updateEnvMap(config, r, xx)

        # finally add the user provided extra env
        self._updateEnvMap(config, r, extra)

        xkeys += extra.keys()
        xkeys = list( set(xkeys) )

        if createEnvFile:
            fileDumper = self._createEnvFileWin32 if config.distribId in ('Windows', ) else self._createEnvFileUnix
            fileDumper(r, xkeys)

        return (r, xkeys)

    def _expandConfigInString(self, item: str, config) -> str:
        if not isinstance(item, str):
            item = str(item)

        values = {
            'libdir': config.libdir,
            'prefix': config.prefix,
            'prefix_posix': config.prefix.as_posix(),
            'prefix_msys2': as_msys2_path(config.prefix),
            'srcdir': self.sourceDir,
            'srcdir_posix': self.sourceDir.as_posix(),
            'builddir': self.buildDir,
            'builddir_posix': self.buildDir.as_posix()
        }
        return item.format(**values)

    def _expandConfigForPath(self, item: str, _config):
        values = {
            'srcdir': self.sourceDir,
            'srcdir_posix': self.sourceDir.as_posix(),
            'builddir': self.buildDir,
            'builddir_posix': self.buildDir.as_posix(),
        }
        return item.format(**values)

    def _expandConfigInlist(self, l: T.List[str], config) -> T.List[str]:
        ret = []
        for item in l:
            ret.append( self._expandConfigInString(item, config) )

        return ret

    def init(self, config) -> bool:
        self.sourceDir = config.sourcesDir / self.name

        dirName = f"{config.targetDistrib}-{config.toolchainObj.description}-{config.targetArch}-{config.buildType}"
        self.buildDir = config.buildsDir / dirName / self.name

        os.makedirs(self.buildDir, exist_ok=True)

        self.logFile = self.buildDir / 'build.log'
        self.prepareStateFile = self.buildDir / PREPARE_DUMP_FILE
        self.builtFile = self.buildDir / BUILT_FILE
        return True

    def checkout(self, _config) -> bool:
        with open(self.logFile, "at", encoding='utf8') as flog:
            return self.srcObj.checkout(self.sourceDir, flog)


    def showLogs(self, header) -> None:
        logging.error(header)
        print (open(self.logFile, "rt", encoding='utf8').read())

    def showLogOnError(self, retcode) -> bool:
        if retcode != 0:
            self.showLogs("error during execute:")
            return False
        return True

    def execute(self, cmd, env, cwd=None) -> bool:
        with open(self.logFile, "at", encoding='utf8') as flog:
            completedProc = subprocess.run(cmd, env=env, cwd=cwd, stdout=flog, stderr=flog)

        return self.showLogOnError(completedProc.returncode)

    def runCommands(self, runItems, env, config) -> bool:
        with open(self.logFile, "at", encoding='utf8') as flog:
            for cmd, path, cmddoc in runItems:
                cmd = self._expandConfigInlist(cmd, config)
                logging.debug(f'{cmddoc}: {" ".join(cmd)}')
                path = pathlib.Path(self._expandConfigForPath(path, config))

                completedProc = subprocess.run(cmd, env=env, cwd=path, stdout=flog, stderr=flog)
                if completedProc.returncode != 0:
                    self.showLogs(f"error {cmddoc} with {' '.join(cmd)}:")
                    return False

        return True

    def _createPrepareFileUnix(self, config, _env, _xkeys) -> None:
        with open(self.buildDir / "prepare.sh", "wt", encoding='utf8') as f:
            f.write(f'# prepare commands for artifact {self.name}\n\n')
            lastDir = None
            for cmd, path, cmddoc in self.prepare_cmds:
                f.write(f'# {cmddoc}\n')

                path = self._expandConfigInString(path, config)
                if lastDir != path:
                    f.write(f'cd "{path}"\n')

                cmd = self._expandConfigInlist(cmd, config)
                f.write('"')
                f.write('" "'.join(cmd))
                f.write('"\n\n')

                lastDir = path

    def _createPrepareFileWin32(self, config, env, xkeys) -> None:
        with open(self.buildDir / WIN_PREPARE_SCRIPT, "wt", encoding='utf8') as f:
            f.write("$PSDefaultParameterValues['*:Encoding'] = 'utf8'\n")

            self._pushPowerShellEnv(f, env, xkeys)

            toolchainContent = config.toolchainObj.prepareItems()
            if toolchainContent:
                f.write(f'# added by toolchain {config.toolchainObj.description}\n')
                f.write(toolchainContent)
                f.write('\n')

            f.write(f'#\n# prepare commands for artifact {self.name}\n#\n\n')
            lastDir = None
            for cmd, path, cmddoc in self.prepare_cmds:
                f.write(f'# {cmddoc}\n')

                path = self._expandConfigInString(path, config)
                if lastDir != path:
                    f.write(f'cd "{path}"\n')

                cmd = self._expandConfigInlist(cmd, config)
                cmdStr = "' '".join(cmd)
                f.write(f"& '{cmdStr}'\nif ($LastExitCode -ne 0) {{\n\tExit $LastExitCode\n}}\n")

                lastDir = path

    def needsRebuildFromDepsUpdates(self, config):
        if not self.builtFile or not os.path.exists(self.builtFile):
            return False

        self_mtime = os.stat(self.builtFile).st_mtime

        for dep in self.deps:
            artifact = config.getBuildItem(dep)
            if artifact and artifact.builtFile and os.path.exists(artifact.builtFile):
                mtime = os.stat(artifact.builtFile).st_mtime
                if self_mtime < mtime:
                    logging.debug(f'rebuilding {self.name} because {artifact.name} has a more recent build')
                    return True
        return False


    def prepare(self, config) -> bool:
        os.makedirs(self.buildDir, exist_ok=True)

        if self.needsRebuildFromDepsUpdates(config):
            # some of our deps have been updated, let's rebuild
            if os.path.exists(self.prepareStateFile):
                os.remove(self.prepareStateFile)

            if os.path.exists(self.builtFile):
                os.remove(self.builtFile)

        (env, xkeys) = self._computeEnv(config, self.extraEnv, config.debug)

        dump = BuildStepDump()
        dump.env = env.copy()
        dump.args = self.prepare_cmds

        dumpOnDisk = None
        if os.path.exists(self.prepareStateFile):
            try:
                with open(self.prepareStateFile, 'rb') as f:
                    dumpOnDisk = pickle.load(f)
            except:
                logging.info(f"prepare state file {self.prepareStateFile} exists but we couldn't read it")

        if dumpOnDisk and dump == dumpOnDisk:
            logging.debug(f"{self.name} is already prepared")
            return True

        if os.path.exists(self.builtFile):
            os.remove(self.builtFile)

        scriptBuilder = self._createPrepareFileWin32 if config.distribId in ('Windows',) else self._createPrepareFileUnix
        scriptBuilder(config, env, xkeys)

        ret = False
        if config.distribId in ('Windows', ) and config.buildWithPowershell:
            logging.debug(f'running powershell .\\{WIN_PREPARE_SCRIPT}')
            ret = self.execute(['powershell', '-ExecutionPolicy', 'Unrestricted', '-File', f'.\\{WIN_PREPARE_SCRIPT}'], env, self.buildDir)
        else:
            ret = self.runCommands(self.prepare_cmds, env, config)

        if ret:
            try:
                with open(self.prepareStateFile, 'wb') as f:
                    pickle.dump(dump, f)
            except Exception as e:
                logging.info(f"unable to save prepare state file {self.prepareStateFile}: {e}")
            return True

        return False

    def createBuiltFile(self) -> bool:
        try:
            with open(self.builtFile, "wt", encoding='utf8') as f:
                f.write('built')

            return True
        except:
            logging.error(f"unable to create built file {self.builtFile} for artifact {self.name}")
            return False

    def _createWin32BuildScript(self, config, env: T.Dict[str, str], xkeys) -> bool:
        with open(self.buildDir / WIN_BUILD_SCRIPT, "wt", encoding='utf8') as f:
            f.write("$PSDefaultParameterValues['*:Encoding'] = 'utf8'\n")

            self._pushPowerShellEnv(f, env, xkeys)

            toolchainItem = config.toolchainObj.prepareItems()
            if toolchainItem:
                f.write(f'# toolchain setup for {config.toolchainObj.description}\n{toolchainItem}\n')

            lastPath = None
            for cmd, path, cmddoc in self.build_cmds:
                f.write(f'# {cmddoc}\n')

                path = pathlib.Path(self._expandConfigForPath(path, config))
                if lastPath != path:
                    f.write(f'cd {path}\n')
                    lastPath = path

                cmd = self._expandConfigInlist(cmd, config)
                cmdStr = "' '".join(cmd)
                f.write(f"& '{cmdStr}'\nif ($LastExitCode -ne 0) {{\n\tExit $LastExitCode\n}}\n")

            return True


    def build(self, config) -> bool:
        if os.path.isfile(self.builtFile):
            logging.debug(f'artifact {self.name} already built')
            return True

        (env, xkeys) = self._computeEnv(config, self.extraEnv)

        if config.distribId in ('Windows',) and config.buildWithPowershell:
            if not self._createWin32BuildScript(config, env, xkeys):
                return False

            logging.debug(f'running powershell .\\{WIN_BUILD_SCRIPT}')
            cmd = ['powershell', '-ExecutionPolicy', 'Unrestricted', '-File', f'.\\{WIN_BUILD_SCRIPT}']
            return self.execute(cmd, env, self.buildDir) and self.createBuiltFile()

        return self.runCommands(self.build_cmds, env, config) and self.createBuiltFile()


    def setMakeNinjaCommands(self, config, cmd='ninja', build_targets='all', install_targets='install', parallelJobs=True,
                            runInstallDir='{builddir}') -> None:
        ''' configure build commands base on make or ninja '''

        if build_targets:
            if isinstance(build_targets, str):
                build_targets = build_targets.split(',')

        if install_targets:
            if isinstance(install_targets, str):
                install_targets = install_targets.split(',')

        if cmd in ('ninja', 'make', 'makeMsys2',):
            maxJobs = 0
            if parallelJobs:
                maxJobs = config.maxJobs

            if maxJobs == 0:
                concurrentArgs = '-j'
            else:
                concurrentArgs = f'-j{maxJobs}'

            if cmd == 'makeMsys2':
                for target in build_targets:
                    self.build_cmds.append( (RunInShell(['make', '-C', '{builddir_posix}', concurrentArgs, target]).expand(),
                            '{builddir_posix}', 'building'))

                for target in install_targets:
                    self.build_cmds.append( (RunInShell(['make', '-C', '{builddir_posix}', concurrentArgs, target]).expand(),
                                  '{builddir_posix}', 'installing'))

            else:
                for target in build_targets:
                    self.build_cmds.append( ([cmd, '-C', '{builddir}', concurrentArgs, target], '{builddir}', 'building'))

                for target in install_targets:
                    self.build_cmds.append(([cmd, '-C', runInstallDir, concurrentArgs, target], '{builddir}', 'installing'))

        elif cmd in ('nmake',):
            for target in build_targets:
                self.build_cmds.append( ([cmd, target], '{builddir}', 'building'))

            for target in install_targets:
                self.build_cmds.append( ([cmd, target], '{builddir}', 'installing') )


class CustomCommandBuildArtifact(BuildArtifact):
    ''' an artifact that is configured with a special command '''

    def __init__(self, name: str, deps, srcObj: Source, extraEnv={}, provides: T.List[str] = [], pkgs={}, prepare_src_cmds=[], prepare_cmds=[],
                 builder='make', build_target='all', install_target='install', toolchainArtifacts='c', skipToolchainEnv=False) -> None:
        '''
            @param name: name of the build artifact
            @param deps: list of dependencies to other build artifacts
            @param srcObj: Source object for checking out code
            @param extraEnv: extra environment variable to use when running commands
            @param provides: list of provided build artifacts
            @param pkgs: required platform packages
            @param prepare_src_cmds: the list of commands to prepare the source directory
            @param prepare_cmds: the list of commands to prepare the build directory
            @param builder: which builder to use can be ninja, make or nmake
            @param build_target: target to build
            @param install_target: install to build
            @param toolchainArtifacts: artifacts that we need from the toolchain
            @param skipToolchainEnv: do not grab env variable from the toolchain
        '''
        BuildArtifact.__init__(self, name, deps, srcObj, extraEnv=extraEnv, provides=provides, pkgs=pkgs,
                               toolchainArtifacts=toolchainArtifacts, skipToolchainEnv=skipToolchainEnv)

        self.builder = builder
        self.build_target = build_target
        self.install_target = install_target

        for cmd in prepare_src_cmds:
            if isinstance(cmd, RunInShell):
                cmd = cmd.expand()
                self.needsMsys2 = True
            self.prepare_cmds.append((cmd, '{srcdir}', f'preparing sources {name}'))

        for cmd in prepare_cmds:
            if isinstance(cmd, RunInShell):
                cmd = cmd.expand()
                self.needsMsys2 = True

            self.prepare_cmds.append((cmd, '{builddir}', f'preparing build tree {name}'))

        if builder == 'makeMsys2':
            self.needsMsys2 = True


    def prepare(self, config) -> bool:
        self.setMakeNinjaCommands(config, self.builder, parallelJobs=self.parallelJobs, build_targets=self.build_target,
                                  install_targets=self.install_target)
        return BuildArtifact.prepare(self, config)


class CMakeBuildArtifact(BuildArtifact):
    ''' cmake based build item '''

    def __init__(self, name: str, deps, srcObj: Source, cmakeOpts=[], parallelJobs=True, extraEnv={}, provides=[], pkgs={},
                 toolchainArtifacts='c') -> None:
        '''
            @param name: name of the build artifact
            @param deps: list of dependencies to other build artifacts
            @param srcObj: Source object for checking out code
            @param extraEnv: extra environment variable to use when running commands
            @param provides: list of provided build artifacts
            @param pkgs: required platform packages
            @param cmakeOpts:
            @param parallelJobs:
            @param toolchainArtifacts: artifacts that we need from the toolchain
        '''
        extra = {
            'Ubuntu|Debian|Redhat|Fedora|Arch|FreeBSD|Darwin': ['cmake']
        }
        doMingwCrossDeps(['Ubuntu', 'Debian', 'Redhat', 'Fedora'], ['cmake', 'ninja-build', 'make'], extra)
        pkgs = mergePkgDeps(pkgs, extra)

        BuildArtifact.__init__(self, name, deps, srcObj, extraEnv, provides, pkgs, toolchainArtifacts=toolchainArtifacts)
        self.cmakeOpts = cmakeOpts
        self.parallelJobs = parallelJobs

    def prepare(self, config) -> bool:
        cmake_cmd = ['cmake']

        if config.crossCompilation:
            fname = config.getCrossPlatformFile("cmake", config.distribId, config.targetDistrib, config.targetArch)
            cmake_cmd.append(f'-DCMAKE_TOOLCHAIN_FILE={fname}')

        #  f'-DCMAKE_BUILD_TYPE={config.cmakeBuildType()}',
        cmake_cmd += [
               '-DCMAKE_PREFIX_PATH={prefix_posix}/lib/cmake;{prefix_posix}/lib',
                f'-DCMAKE_CONFIGURATION_TYPES={config.cmakeBuildType()}',
               '-DCMAKE_INSTALL_PREFIX={prefix_posix}',
               '-S', '{srcdir}',
               '-B', '{builddir}'
        ]

        cmake_cmd += self.cmakeOpts

        self.prepare_cmds = [
            (cmake_cmd, '{builddir}', 'running cmake')
        ]

        self.build_cmds = [
            (['cmake', '--build', '{builddir}', '--config', config.cmakeBuildType()], '{builddir}', 'building'),
            (['cmake', '--install', '{builddir}'], '{builddir}', 'installing'),
        ]
        return BuildArtifact.prepare(self, config)



class QMakeBuildArtifact(BuildArtifact):
    ''' qmake + make based build item '''

    def __init__(self, name, deps, srcObj: Source, extraEnv={}, provides=[], pkgs={}, toolchainArtifacts='c++') -> None:
        '''
            @param name: name of the build artifact
            @param deps: list of dependencies to other build artifacts
            @param srcObj: Source object for checking out code
            @param extraEnv: extra environment variable to use when running commands
            @param provides: list of provided build artifacts
            @param pkgs: required platform packages
            @param toolchainArtifacts: artifacts that we need from the toolchain
        '''
        pkgs = mergePkgDeps(pkgs, {
            'Ubuntu|Debian': ['qt5-qmake'],
            'Redhat|Fedora': ['qt5-qtbase-devel'],
            'FreeBSD': ['qt5-qmake'],
        })

        BuildArtifact.__init__(self, name, deps, srcObj, extraEnv, provides, pkgs, toolchainArtifacts=toolchainArtifacts)

        # needed to avoid some g++ link errors on Fedora
        self.extraEnv = extraEnv.copy()

        self.extraEnv.update({
            'RPM_ARCH': 'bla',
            'RPM_PACKAGE_RELEASE': 'bla',
            'RPM_PACKAGE_VERSION': 'bla',
            'RPM_PACKAGE_NAME': 'bla'
        })

    def prepare(self, config) -> bool:
        cmd = []
        qtChooser = config.distribId in ('Ubuntu', 'Debian', )
        if qtChooser:
            cmd = ['qtchooser', '-qt=qt5', '--run-tool=qmake']
        else:
            cmd = ['qmake-qt5']

        cmd += [
            'ADDITIONAL_RPATHS={prefix}/{libdir}/',
            'PREFIX={prefix}',
            '{srcdir}'
        ]

        self.prepare_cmds = [
            (cmd, '{builddir}', 'running qmake')
        ]

        self.setMakeNinjaCommands(config, 'make', parallelJobs=self.parallelJobs)
        return BuildArtifact.prepare(self, config)



class AutogenBuildArtifact(BuildArtifact):
    ''' autotools/autogen.sh + make based build item '''

    def __init__(self, name, deps, srcObj: Source, extraEnv={}, provides=[], pkgs={}, autogenArgs=[], noconfigure=False,
                 isAutogen=True, bootstrapScript='bootstrap.sh', configureArgs=[], runInstallDir='{builddir}', toolchainArtifacts='c') -> None:
        '''
            @param name: name of the build artifact
            @param deps: list of dependencies to other build artifacts
            @param srcObj: Source object for checking out code
            @param extraEnv: extra environment variable to use when running commands
            @param provides: list of provided build artifacts
            @param pkgs: required platform packages
            @param autogenArgs:
            @param noconfigure:
            @param isAutogen:
            @param bootstrapScript:
            @param configureArgs:
            @param runInstallDir:
            @param toolchainArtifacts: artifacts that we need from the toolchain
        '''
        pkgs = mergePkgDeps(pkgs, {
            'Ubuntu|Debian': ['make', 'build-essential', 'automake', 'autoconf', 'libtool'],
            'Fedora|Redhat': ['make', 'autoconf', 'automake', 'libtool'],
            'Arch|Darwin|FreeBSD': ['autoconf', 'automake', 'libtool'],
        })

        BuildArtifact.__init__(self, name, deps, srcObj, extraEnv, provides, pkgs, toolchainArtifacts=toolchainArtifacts)

        self.isAutogen = isAutogen
        self.autogenArgs = autogenArgs
        self.noconfigure = noconfigure
        self.configureArgs = configureArgs
        self.bootstrapScript = bootstrapScript
        self.runInstallDir = runInstallDir

    def prepare(self, config) -> bool:
        if self.isAutogen:
            cmd = [os.path.join(self.sourceDir, "autogen.sh")] + self.autogenArgs
        else:
            cmd = [os.path.join(self.sourceDir, self.bootstrapScript)]

        autoGenRunDir = '{srcdir}'
        if self.noconfigure:
            cmd += ["--prefix={prefix}"]
            autoGenRunDir = '{builddir}'

        self.prepare_cmds = [
            (cmd, autoGenRunDir, 'running autogen/bootstrap')
        ]

        if not self.noconfigure:
            cmd = [ os.path.join(self.sourceDir, "configure"), "--prefix={prefix}"] + self.configureArgs
            self.prepare_cmds.append(
                (cmd, '{builddir}', 'running configure')
            )

        self.setMakeNinjaCommands(config, 'make', parallelJobs=self.parallelJobs, runInstallDir=self.runInstallDir)
        return BuildArtifact.prepare(self, config)


class MesonBuildArtifact(BuildArtifact):
    ''' meson + ninja based build item '''

    def __init__(self, name, deps, srcObj: Source, extraEnv={}, provides=[], pkgs={}, mesonOpts=[], parallelJobs=True,
                 toolchainArtifacts='c') -> None:
        '''
            @param name: name of the build artifact
            @param deps: list of dependencies to other build artifacts
            @param srcObj: Source object for checking out code
            @param extraEnv: extra environment variable to use when running commands
            @param provides: list of provided build artifacts
            @param pkgs: required platform packages
            @param mesonOpts:
            @param parallelJobs:
            @param toolchainArtifacts: artifacts that we need from the toolchain
        '''
        extra = {
            'Ubuntu|Debian|Redhat|Fedora': ['meson', 'ninja-build'],
            'Arch|FreeBSD|Darwin': ['meson', 'ninja'],
        }
        doMingwCrossDeps(['Ubuntu', 'Debian', 'Redhat', 'Fedora'], ['meson', 'ninja-build'], extra)

        pkgs = mergePkgDeps(pkgs, extra)
        BuildArtifact.__init__(self, name, deps, srcObj, extraEnv, provides, pkgs, toolchainArtifacts=toolchainArtifacts)
        self.mesonOpts = mesonOpts
        self.parallelJobs = parallelJobs

    def prepare(self, config) -> bool:
        reconfigure = os.path.exists(self.buildDir / 'meson-info')

        cmd = ['meson', 'setup',
               '-Dprefix={prefix}',
               f'-Dbuildtype={config.mesonBuildType()}',
        ]

        if config.crossCompilation:
            cmd += ['--cross-file', config.getCrossPlatformFile('meson',  config.distribId, config.targetDistrib, config.targetArch)]

        if reconfigure:
            cmd += ["--reconfigure"]
        cmd += self.mesonOpts
        cmd += [self.sourceDir]

        self.prepare_cmds = [
            (cmd, '{builddir}', 'running meson configure')
        ]

        maxJobs = 0
        if self.parallelJobs:
            maxJobs = config.maxJobs

        concurrentArgs = []
        if maxJobs != 0:
            concurrentArgs = ['-j', f'{maxJobs}']

        self.build_cmds = [
            (['meson', 'compile'] + concurrentArgs, '{builddir}', 'building'),
            (['meson', 'install'], '{builddir}', 'installing'),
        ]

        return BuildArtifact.prepare(self, config)



import os
import pathlib
import typing as T
from packaging.version import Version

from zenlog import log as logging
from accendino.platform import accendinoPlatform

def checkAccendinoVersion(cond: str, v: str) -> bool:
    ''' '''
    tokens = cond.split(' ', 2)
    if len(tokens) != 2:
        logging.error(f"invalid condition string {cond}")
        return False

    op = tokens[0]
    checkVersion = tokens[1]

    if op in ('=', '==',):
        return checkVersion == v

    if op in ('!', '!=',):
        return checkVersion != v

    if op == '<':
        return Version(v) < Version(checkVersion)

    if op == '<=':
        return Version(v) <= Version(checkVersion)

    if op == '>':
        return Version(v) > Version(checkVersion)

    if op == '>=':
        return Version(v) >= Version(checkVersion)

    logging.error(f"operation {op} not supported")
    return False


def checkVersionCondition(cond: str, distribId: str, distribVersion: str) -> bool:
    ''' '''
    tokens = cond.split(' ', 3)
    if len(tokens) < 2:
        logging.error(f"invalid condition string {cond}")
        return False

    op = tokens[0]
    operSys = tokens[1]

    if op in ('=', '==',):
        if distribId != operSys:
            return False

        if len(tokens) < 3:
            return True
        return distribVersion == tokens[2]

    if op in ('!', '!=',):
        if distribId == operSys:
            return False

        if len(tokens) < 3:
            return True
        return distribVersion != tokens[2]

    # do version comparisons from here
    if distribId != operSys:
        return False

    if len(tokens) < 3:
        logging.error(f"invalid condition string {cond}")
        return False

    version = tokens[2]
    if op == '<':
        return distribVersion < version

    if op == '<=':
        return distribVersion <= version

    if op == '>':
        return distribVersion > version

    if op == '>=':
        return distribVersion >= version

    logging.error(f"operation {op} not supported")
    return False

class ConditionalDep:
    ''' '''
    def __init__(self, cond) -> None:
        ''' '''
        self.conditions = cond.strip().split('|')

    def apply(self, _config, deps: T.List[str]) -> T.List[str]:
        ''' '''
        return deps

    def condVerified(self, config) -> bool:
        for cond in self.conditions:
            if checkVersionCondition(cond, config.distribId, config.distribVersion):
                return True
        return False

class DepsAdjuster(ConditionalDep):
    ''' '''

    def __init__(self, cond, add=[], drop=[]) -> None:
        ''' '''
        ConditionalDep.__init__(self, cond)
        self.addItems = add
        self.dropItems = drop

    def apply(self, config, deps: T.List[str]) -> T.List[str]:
        ''' '''
        if not self.condVerified(config):
            return deps

        ret = deps[:]
        ret += self.addItems
        for i in self.dropItems:
            ret.remove(i)
        return ret


def treatPackageDeps(pkgs: T.Dict[str, T.Any]) -> T.Dict[str, T.Any]:
    ''' expands a package deps map, it treats keys that have the form
        `'k1|k2|k3': V` by splitting and setting V for the k1, k2, k3 keys
        Note that we expect V to be a list and so you can have concatenation by
        doing:
            'k1|k2': V1
            'k1': V2

        You will end up with
            'k1': V1 + V2,
            'k2': V1
    '''
    ret = {}

    for k, v in pkgs.items():
        keys = k.split('|')
        for key in keys:
            baseV = ret.get(key, [])
            ret[key] = baseV + v[:]

    return ret

def mergePkgDeps(d1: T.Dict[str, T.Any], d2: T.Dict[str, T.Any]) -> T.Dict[str, T.Any]:
    ret = d1.copy()
    for k, v in d2.items():
        d1_value = ret.get(k, [])
        ret[k] = d1_value + v

    return ret


def doMingwCrossDeps(distribs: T.List[str], deps: T.List[T.Any], target: T.Dict[str, T.Any]):
    for d in distribs:
        for arch in ('x86_64', 'i686',):
            key = f'{d}->mingw@{arch}'
            baseV = target.get(key, [])
            target[key] = baseV + deps[:]


class NativePath:
    ''' class that represents a native path '''
    def __init__(self, *args, **kwargs):
        self.prefix = kwargs.get('prefix', '')
        self.suffix = kwargs.get('suffix', '')
        self.items = args[:]

    def __str__(self) -> str:
        return self.prefix + str(pathlib.PurePath(*self.items)) + self.suffix

class RunInShell:
    ''' class that represent something to execute in a shell (UNIX shell or msys2) '''
    def __init__(self, args):
        self.items = args[:]

    def expand(self):
        if accendinoPlatform.isWindows:
            cmdItems = []
            for item in self.items:
                cmdItems.append(str(item))

            return [accendinoPlatform.msys2path, '-defterm', '-no-start', '-mingw64', '-here', '-c', ' '.join(cmdItems)]

        return self.items

def findInPATH(name: str) -> str:
    if accendinoPlatform.isWindows:
        if not name.endswith(".exe"):
            name += ".exe"

    for p in os.environ.get('PATH', '').split(os.pathsep):
        fpath = os.path.join(p, name)
        if os.path.isfile(fpath) and os.access(fpath, os.X_OK):
            return fpath
    return None

def escapeForPowershell(s):
    ret = ''
    haveSpace = False

    for c in s:
        if c == '\\':
            c = '\\\\'
        elif c == ' ':
            haveSpace = True

        ret += c

    if haveSpace:
        return  "'" + ret + "'"

    return ret

def as_msys2_path(p) -> str:
    if p.drive:
        l = [ p.drive[:-1].lower() ] + list( p.parts[1:] )
        return '/' + '/'.join(l)

    return p.as_posix()

def is_exact_instance(obj, klass):
    s1 = obj.__class__.__name__
    s2 = klass.__name__
    return s1 == s2

def getArchLibDir(distribId: str, archId: str) -> str:
    if distribId == 'mingw':
        return None

    OS_KEYS = {
        'Ubuntu': '-linux-gnu',
        'Debian': '-linux-gnu',
        'Fedora': '-linux-gnu',
        'Redhat': '-linux-gnu'
    }

    osKey = OS_KEYS.get(distribId, distribId)
    ARCH_KEYS = {
        'amd64': 'x86_64',
        'x86_64': 'x86_64',
    }

    archKey = ARCH_KEYS.get(archId, None)
    if not archKey:
        logging.error("arch {0} not handled yet", archId)
    return archKey + osKey


def envForDepotTools(env: T.Dict[str, str] = None):
    ''' cleanups PYTHONPATH for gclient, fetch and gn otherwise we have conflicts between
        python versions
    '''
    if env is None:
        env = os.environ.copy()
    else:
        env = env.copy()

    if 'PYTHONPATH' in env:
        del env['PYTHONPATH']

    return env



#! /usr/bin/env python3
import sys
import getopt
import os.path
import platform
import pathlib
import typing as T
import configparser
from zenlog import log as logging

import accendino

from accendino.builditems import BuildArtifact, AutogenBuildArtifact, CMakeBuildArtifact, DepsBuildArtifact, \
    MesonBuildArtifact, QMakeBuildArtifact, CustomCommandBuildArtifact
from accendino.localdeps import getPkgManager
from accendino.sources import LocalSource, GitSource, RemoteArchiveSource
from accendino.utils import ConditionalDep, DepsAdjuster, checkVersionCondition, checkAccendinoVersion, \
    NativePath, RunInShell, mergePkgDeps, is_exact_instance
from accendino.toolchain import getToolchain



def doHelp(args, is_error) -> int:
    ''' '''
    print(f"usage: {args[0]} [--help] [--prefix=<prefix>] [--debug] [--targets=<targets>] [--build-type=<type>] <file>")
    print("\t--help: shows this help")
    print("\t--version: prints the version")
    print("\t--debug: show verbose information of building")
    print("\t--prefix=<prefix>: where to install")
    print("\t--targets=<targets>: a list of comma separated targets to build (by default full-ogon-freerdp2)")
    print("\t--build-type=[release|debug]: type of build (defaults to release)")
    print("\t--work-dir=<path>: a path to the working directory where sources are checked out and built")
    print("\t--options=<path>: a path to the build options file")
    if is_error:
        return 1

    return 0

def detectPlatform() -> None:
    ''' '''
    distrib_id = None
    distrib_version = None

    plat_system = platform.system()
    if plat_system == 'Linux':
        # try reading LSB file
        try:
            for l in open("/etc/lsb-release", "rt", encoding='utf8').readlines():
                pos = l.find("DISTRIB_ID=")
                if pos == 0:
                    distrib_id = l[len('DISTRIB_ID='):-1]
                    continue

                pos = l.find('DISTRIB_RELEASE=')
                if pos == 0:
                    distrib_version = l[len('DISTRIB_RELEASE='):-1]
                    continue

            return (distrib_id, distrib_version)
        except:
            logging.debug(" * failed to detect platform using /etc/lsb-release (not an error)")

        # Debian based systems
        try:
            content = open("/etc/debian_version", "rt", encoding='utf8').readline()
            distrib_id = "Debian"
            tokens = content.split('.', 2)
            distrib_version = f"{int(tokens[0]):02d}.{int(tokens[1]):02d}"
            return (distrib_id, distrib_version)
        except:
            logging.debug(" * failed reading /etc/debian_version  (not an error)")

        # Fedora based systems
        try:
            content = open("/etc/fedora-release", "rt", encoding='utf8').readline()
            distrib_id = "Fedora"
            tokens = content.split(' ', 4)
            distrib_version = tokens[2]
            return (distrib_id, distrib_version)
        except:
            logging.debug(" * failed reading /etc/fedora-version  (not an error)")

    elif plat_system in ('Linux', 'FreeBSD',):
        # Arch / FreeBSD
        try:
            distrib_id = None
            distrib_ver = None
            for l in open("/etc/os-release", "rt", encoding='utf8').readlines():
                l = l.strip()
                if l.startswith('ID='):
                    distrib_id = l[3:]
                elif l.starstwith('BUILD_ID='):
                    distrib_ver = l[9:]
                elif l.startswith('VERSION_ID="'):
                    distrib_ver = l[len('VERSION_ID="') : -1]

                if distrib_id is not None and distrib_ver is not None:
                    return (distrib_id, distrib_ver)
        except:
            logging.debug(" * failed reading /etc/os-release")

    elif plat_system in ("Windows", "Darwin",):
        distrib_id = plat_system
        if plat_system == "Windows":
            distrib_version = platform.win32_ver()[0]
        else:
            distrib_version = platform.mac_ver()[0]

    return (distrib_id, distrib_version)


BUILD_TYPES = ('release', 'debug',)
ARCHS = ('i686', 'x86_64')

ARCHS_MAP = {
    'AMD64': 'x86_64', # Windows
    'amd64': 'x86_64', # FreeBSD
}

class AccendinoConfig:
    ''' Accendino configuration '''

    def __init__(self) -> None:
        ''' '''
        self.debug = False
        self.buildWithPowershell = True
        self.buildType = 'release'
        self.projectName = None
        self.workDir = pathlib.PurePath(os.getcwd())
        self.prefix = None
        self.projectDir = None
        self.sourcesDir = None
        self.buildsDir = None
        self.toolsDir = None
        self.targets = None
        self.buildDefs = []
        self.distribId = None
        self.distribVersion = None
        self.checkPackages = True
        self.doBuild = True
        self.libdir = 'lib'
        self.resumeFrom = None
        self.maxJobs = 5
        self.crossCompilation = False
        self.toolchain = 'default'
        self.toolchainObj = None

        localMachine = platform.machine()
        self.localArch = ARCHS_MAP.get(localMachine, localMachine)
        self.targetArch = None
        self.targetDistrib = None
        self.pocketDir = pathlib.Path(__file__).parent / 'pocket'
        self.crossFileDir = pathlib.Path(__file__).parent / 'cross'
        self.getCrossPlatformFile = None
        self.optionsFile = None
        self.options = None
        self.includesStack = []

        #
        # populate a search path that has:
        #   * first paths provided in ACCENDINO_PATH,
        #   * then in the accendino pocket
        self.pocketSearchPaths = os.environ.get('ACCENDINO_PATH', '').split(':') + [ self.pocketDir ]
        self.includedFiles = []

        def includeFn(fname: str, include_once: bool = True) -> bool:
            ''' '''
            if not fname.endswith('.accendino'):
                fname = fname + '.accendino'

            logging.debug(f"including file '{fname}'")
            ret = self.readSource(fname, include_once)
            if ret:
                self.includedFiles.append(fname)
            return ret


        def pickDeps(name: str) -> T.List[T.Any]:
            ''' '''
            targets = self.context.get("ARTIFACTS", [])
            for t in targets:
                if t.name == name:
                    return t.deps[:]
            return None

        def pickPkgDeps(name: str, extra = None, override : bool = False) -> T.Dict[str, T.List[str]]:
            ''' '''
            ret = None
            targets = self.context.get("ARTIFACTS", [])
            for t in targets:
                if t.name == name:
                    ret = t.pkgs.copy()
                    break

            if ret is None:
                logging.error(f'error picking package dependencies of {name}')

            if extra:
                for k, v in extra.items():
                    for subk in k.split('|'):
                        if override:
                            base_v = []
                        else:
                            base_v = ret.get(subk, [])

                        ret[subk] = base_v + v

            return ret

        def checkDistrib(cond: str) -> bool:
            return checkVersionCondition(cond, self.distribId, self.distribVersion)

        def checkAccendinoVersionFn(cond: str) -> bool:
            return checkAccendinoVersion(cond, accendino.__version__)

        def getOption(optName: str, defaultVal=None):
            if self.options is None:
                return defaultVal

            tokens = optName.split('.', 2)
            if len(tokens) == 2:
                (section, name) = tokens
            else:
                section = configparser.UNNAMED_SECTION
                name = optName

            if not section in self.options.sections():
                return defaultVal

            ret = self.options[section].get(name, defaultVal)
            if isinstance(ret, str):
                if ret.lower() in ('true', 'on', 'yes',):
                    return True

                if ret.lower() in ('false', 'off', 'no',):
                    return False

            return ret


        self.sources = []
        self.context = {
            'ARTIFACTS': [],
            'DEFAULT_TARGETS': None,
            'LocalSource': LocalSource,
            'GitSource': GitSource,
            'RemoteArchiveSource': RemoteArchiveSource,
            'DepsBuildArtifact': DepsBuildArtifact,
            'CMakeBuildArtifact': CMakeBuildArtifact,
            'QMakeBuildArtifact': QMakeBuildArtifact,
            'AutogenBuildArtifact': AutogenBuildArtifact,
            'MesonBuildArtifact': MesonBuildArtifact,
            'CustomCommandBuildArtifact': CustomCommandBuildArtifact,
            'BuildArtifact': BuildArtifact,
            'include': includeFn,
            'getOption': getOption,
            'checkDistrib': checkDistrib,
            'pickDeps': pickDeps,
            'pickPkgDeps': pickPkgDeps,
            'mergePkgDeps': mergePkgDeps,
            'logging': logging,
            'DepsAdjuster': DepsAdjuster,
            'NativePath': NativePath,
            'RunInShell': RunInShell,
            'accendinoVersion': accendino.__version__,
            'UBUNTU_LIKE': 'Debian|Ubuntu',
            'REDHAT_LIKE': 'Fedora|Redhat',
            'checkAccendinoVersion': checkAccendinoVersionFn,
        }

    def findSourceFile(self, fname: str, include_once: bool = True) -> str:
        searchPaths = ['.', 'pocket'] + self.pocketSearchPaths

        # also search in the directory of the last included file
        if len(self.includesStack) > 0:
            searchPaths.append( os.path.dirname(self.includesStack[-1]) )

        # if the requested file is a path then just try to open it
        if fname.startswith('.') or fname.startswith('/'):
            searchPaths = ['']

        for p in searchPaths:
            fpath = os.path.join(p, fname)
            if os.path.exists(fpath) and os.path.isfile(fpath):
                if include_once and fpath in self.includedFiles:
                    logging.debug(f"file '{fpath}' already included")
                    return True
                return fpath

        return None

    def cmakeBuildType(self) -> str:
        ''' '''
        if self.buildType == 'release':
            return 'Release'

        if self.buildType == 'debug':
            return 'Debug'

        raise Exception(f"{self.buildType} build type not supported for cmake")

    def mesonBuildType(self) -> str:
        ''' '''
        if self.buildType in ['release', 'debug']:
            return self.buildType
        raise Exception(f"{self.buildType} build type not supported for meson")

    def getBuildItem(self, name: str) -> BuildArtifact:
        ''' '''
        for item in self.buildDefs:
            if item.name == name or name in item.provides:
                return item
        return None

    def default_getCrossPlatformFile(self, builder: str, localDistrib: str, targetDistrib : str, arch: str):
        ''' '''
        if targetDistrib == 'mingw':
            bits = '64' if arch == 'x86_64' else '32'

            if localDistrib in ('Redhat', 'Fedora',):
                return f'/usr/share/mingw/toolchain-mingw{bits}.{builder}'

            if localDistrib in ('Ubuntu', 'Debian',):
                return str(self.crossFileDir / f'debian-mingw{bits}.{builder}')

        logging.error(f'no cross platform file available for {targetDistrib}/{arch}')
        return None

    def setPlatform(self, distribId: str, distribVersion: str) -> None:
        self.distribId = distribId
        self.distribVersion = distribVersion
        if not self.targetDistrib:
            self.targetDistrib = distribId

        self.crossCompilation = (self.distribId != self.targetDistrib) or (self.localArch != self.targetArch)

        if self.distribId in ['Redhat', 'Fedora']:
            self.libdir = 'lib64'

        extraKeys = {
            'distribId': distribId,
            'distribVersion': distribVersion,
            'targetArch': self.targetArch,
            'targetDistribId': self.targetDistrib,
            'crossCompilation': self.crossCompilation,
            'buildType': self.buildType,
            'toolchain': self.toolchain,
            'libdir': self.libdir,
        }

        self.context.update(extraKeys)


    def treatPlatformPackages(self, pkgManager, packagesToCheck, buildItems) -> None:
        ''' '''
        # let's compute platform package requirements
        shortName = f"{self.distribId}"
        longName = f"{self.distribId} {self.distribVersion}"
        if self.targetDistrib != self.distribId:
            # cross compiling adjusting the searched names
            shortName += f'->{self.targetDistrib}@{self.targetArch}'
            longName += f'->{self.targetDistrib}@{self.targetArch}'

        toolchainArtifacts = []
        for item in buildItems:
            if isinstance(item, str):
                continue

            toolchainArtifacts += item.toolchainArtifacts

            if longName in item.pkgs:
                pkgs = item.pkgs[longName]
            elif shortName in item.pkgs:
                pkgs = item.pkgs[shortName]
            else:
                if not self.crossCompilation:
                    logging.debug(f" * warning, {item.name} has no package dependency for the current platform")
                continue

            packagesToCheck += pkgs[:]

        packagesToCheck = list(set(packagesToCheck))
        toolchainArtifacts = list(set(toolchainArtifacts))

        if pkgManager:
            if not self.toolchainObj.packagesCheck(pkgManager, toolchainArtifacts, True):
                logging.error(" * package requirements not met")
                return 5

            toInstall = pkgManager.checkMissing(packagesToCheck)
            if toInstall is None:
                logging.error(" * package requirements not met")
                return 5

            if toInstall:
                if not pkgManager.installPackages(toInstall):
                    logging.error(" * error during package installation")
                    return 4

        return 0

    def createBuildPlan(self, itemsToBuild, buildPlan) -> None:
        ''' '''

        def addBuildItems(items: T.List[str], buildPlan: T.List[str], provided: T.List[str]):
            for item in items:
                if item in buildPlan + provided:
                    continue

                itemObj = self.getBuildItem(item)
                if itemObj:
                    if itemObj.deps:
                        addBuildItems(itemObj.deps, buildPlan, provided)

                    buildPlan.append(itemObj.name)
                    if itemObj.provides:
                        if isinstance(itemObj.provides, list):
                            provided += itemObj.provides
                        else:
                            provided.append(itemObj.provides)

                else:
                    logging.error(f"unable to find build dependency {item}")

        plan = []
        provided = []
        addBuildItems(itemsToBuild, plan, provided)

        for itemStr in plan:
            item = self.getBuildItem(itemStr)
            if item:
                buildPlan.append(item)
            else:
                buildPlan.append(itemStr)


    def readSource(self, fname: str, include_once: bool) -> bool:
        ''' '''
        fpath = self.findSourceFile(fname, include_once)
        if isinstance(fpath, bool):
            # already included
            return True

        if fpath is None:
            logging.error(f'file {fname} not found')
            sys.exit(-1)

        self.includesStack.append(fpath)
        with open(fpath, "rt", encoding="utf8") as f:
            code = compile(f.read(), os.path.basename(fpath), "exec")
            exec(code, {}, self.context)
        self.includesStack.pop()
        return True

    def finalizeConfig(self) -> bool:
        ''' '''
        #
        for buildItem in self.context.get('ARTIFACTS', []):
            newDeps = []
            for dep in buildItem.deps:
                if isinstance(dep, str):
                    newDeps.append(dep)
                elif isinstance(dep, ConditionalDep):
                    newDeps = dep.apply(self, newDeps)
                else:
                    logging.error(f'don\'t know how to treat artifact dependency with type {type(dep)}')
            buildItem.deps = newDeps

            newPkg = {}
            for k, v in buildItem.pkgs.items():
                newList = []
                for dep in v:
                    if isinstance(dep, str):
                        newList.append(dep)
                    elif isinstance(dep, ConditionalDep):
                        newList = dep.apply(self, newList)
                    else:
                        logging.error(f'don\'t know how to treat package dependency with type {type(dep)}')

                newPkg[k] = newList
            buildItem.pkgs = newPkg

            self.buildDefs.append(buildItem)

        if not self.targets:
            defaultTargets = self.context.get('DEFAULT_TARGETS', 'ogon')
            if defaultTargets:
                self.targets = defaultTargets.split(",")

        if not self.projectName:
            self.projectName = self.context.get('PROJECT', 'work')

        self.projectDir = pathlib.PurePath(self.workDir, self.projectName)
        if not self.prefix:
            self.prefix = self.projectDir / 'deploy'

        self.sourcesDir = self.projectDir / 'sources'
        self.buildsDir = self.projectDir / 'build'
        self.toolsDir = self.projectDir / 'tools'

        self.getCrossPlatformFile = self.context.get('CROSS_PLATFORM_FILE_CHOOSER', self.default_getCrossPlatformFile)
        return True


def createWorkTree(config) -> int:
    ''' '''
    dirs = [
        (config.prefix, 'root directory'),
        (config.sourcesDir, 'sources directory'),
        (config.buildsDir, 'builds directory'),
    ]

    for dirItem in dirs:
        if not os.path.exists(dirItem[0]):
            logging.debug(f" * creating {dirItem[1]} {dirItem[0]}")
            os.makedirs(dirItem[0])
        else:
            logging.debug(f" * {dirItem[1]} {dirItem[0]} already exists")

        if not os.path.isdir(dirItem[0]):
            logging.error(f"{dirItem[1]} {dirItem[0]} is not a directory")
            return 1

    return 0

_ARGS_CONTINUE, _ARGS_HELP, _ARGS_VERSION, _ARGS_ERROR = range(4)

def treatArgOrOption(config, option, value, fromCmdLine) -> int:
    if not fromCmdLine:
        option = '--' + option

    if option in ('-h', '--help',):
        return _ARGS_HELP

    if option in ('-v', '--version',):
        print(accendino.__version__)
        return _ARGS_VERSION

    if option in ('--prefix',):
        config.prefix = pathlib.PurePath(os.path.abspath(value))
    elif option in ('-d', '--debug',):
        logging.level('debug')
        config.debug = True
    elif option in ('--build-type',):
        config.buildType = value
        if config.buildType not in BUILD_TYPES:
            print(f"invalid build type {config.buildType}")
            return _ARGS_ERROR
    elif option in ('--targets',):
        config.targets = value.split(',')
    elif option in ('--buildWithPowershell',):
        config.buildWithPowershell = True
    elif option in ('--no-packages',):
        config.checkPackages = False
    elif option in ('--build-deps',):
        config.doBuild = False
    elif option in ('--work-dir', ):
        config.workDir = pathlib.PurePath(os.path.abspath(value))
    elif option in ("--resume-from", ):
        config.resumeFrom = value
    elif option in ('--options',):
        if not fromCmdLine:
            logging.error("options file can't be given in an options file")
            return _ARGS_ERROR
        config.optionsFile = value
    elif option in ("--targetArch", ):
        if not value in ARCHS:
            logging.error(f'arch {value} not yet supported')
            return _ARGS_ERROR
        config.targetArch = value
    elif option in ("--targetDistrib", ):
        if value == 'mingw32':
            config.targetDistrib = 'mingw'
            config.targetArch = 'i686'
            config.toolchain = 'mingw'
        elif value == 'mingw64':
            config.targetDistrib = 'mingw'
            config.targetArch = 'x86_64'
            config.toolchain = 'mingw'
        else:
            config.targetDistrib = value
    elif option in ('--toolchain',):
        config.toolchain = value
    elif option in ("--project", ):
        config.projectName = value
    else:
        logging.error(f"unknown option {value}")
        return _ARGS_ERROR

    return _ARGS_CONTINUE


def run(args: T.List[str]) -> int:
    ''' '''
    logging.level("info")

    config = AccendinoConfig()

    opts, extraArgs = getopt.getopt(args[1:], "hdv", [
        "prefix=", "help", "debug", "no-packages", "build-deps", "targets=", "build-type=", "options=",
        "work-dir=", "resume-from=", "project=", "targetDistrib=", "targetArch=", "toolchain=",
        "buildWithPowershell", "version"
    ])

    for option, value in opts:
        argRes = treatArgOrOption(config, option, value, True)
        if argRes in (_ARGS_ERROR, _ARGS_HELP,):
            return doHelp(args[1:], argRes == _ARGS_ERROR)
        if argRes == _ARGS_VERSION:
            return 0

    logging.info("=== accendino ===")

    if config.optionsFile:
        logging.debug(f'reading build options from {config.optionsFile}')
        config.options = configparser.ConfigParser()
        config.options.optionxform = str
        config.options.read(config.optionsFile)

        if 'accendino' in config.options.sections():
            # inject key/values in the accendino section as command line args
            for option, value in config.options['accendino'].items():
                argRes = treatArgOrOption(config, option, value, False)
                if argRes in (_ARGS_ERROR, _ARGS_HELP, _ARGS_VERSION):
                    logging.error(f'invalid item {option}={value} in accendino section')
                    return 1


    config.sources += extraArgs
    if not config.sources: # defaults to build ogon
        logging.info('no source file provided using default ogon.accendino')
        config.sources += ['ogon.accendino']

    (distribId, distribVersion) = detectPlatform()

    if config.targetArch is None:
        config.targetArch = config.localArch

    logging.debug(f"target installation: {distribId} {distribVersion}")
    config.setPlatform(distribId, distribVersion)

    # grab a toolchain
    config.toolchainObj = getToolchain(config.toolchain, config)
    if not config.toolchainObj:
        logging.error(f"unable to find toolchain {config.toolchain}")
        return 2

    for f in config.sources:
        if not config.readSource(f, True):
            logging.error(f"error interpreting {f}")
            return 2
    config.finalizeConfig()

    retCode = createWorkTree(config)
    if retCode:
        return retCode

    buildList = []
    for t in config.targets:
        item = config.getBuildItem(t)
        if item:
            buildList.append(item)

    buildPlan = []
    config.createBuildPlan(config.targets, buildPlan)
    if config.debug:
        items = []
        for i in buildPlan:
            items.append(i.name)

        logging.debug(f"build plan: [{', '.join(items)}]")

    if config.checkPackages:
        packagesToCheck = []
        pkgManager = getPkgManager(config.distribId, packagesToCheck)

        retCode = config.treatPlatformPackages(pkgManager, packagesToCheck, buildPlan)
        if retCode:
            return retCode

    logging.debug(f'==> activating toolchain {config.toolchainObj.description}')
    if not config.toolchainObj.activate():
        logging.error('error activating toolchain')
        return 6


    def buildModule(buildItem) -> bool:
        ''' '''
        logging.debug('==> preparing')
        if not buildItem.prepare(config):
            logging.error(f"prepare error for {buildItem.name}")
            return False

        logging.debug('==> building')
        if not buildItem.build(config):
            logging.error(f"build error for {buildItem.name}, check logs in {buildItem.logFile}")
            return False
        return True

    exitCode = 0
    if config.doBuild:
        for item in buildPlan:
            isDepsBuildArtifact = is_exact_instance(item, DepsBuildArtifact)

            extra = ' is only deps' if isDepsBuildArtifact else ''
            logging.info(f' * module {item.name}{extra}')

            if not isDepsBuildArtifact:
                logging.debug('==> init')
                if not item.init(config):
                    logging.error(f"error initializing {item.name}")
                    return 1

                logging.debug('==> checking out')
                if not item.checkout(config):
                    logging.error(f"checkout error for {item.name}")
                    return 1

            if config.resumeFrom:
                if item.name != config.resumeFrom:
                    continue

                # We've found the item to resume the build from
                config.resumeFrom = None

            if not isDepsBuildArtifact:
                if not buildModule(item):
                    return 1

    logging.info("=== finished ===")
    return exitCode

def main() -> int:
    ''' '''
    return run(sys.argv)

if __name__ == '__main__':
    if platform.python_version_tuple()[0] != '3':
        print("expecting python 3, exiting")
        sys.exit(1)

    sys.exit( main() )
